"""add preopen ports

Revision ID: 3bb80d1887d6
Revises: 1e8531583e20
Create Date: 2020-02-05 17:02:42.344726

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3bb80d1887d6"
down_revision = "1e8531583e20"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("kernels", sa.Column("preopen_ports", sa.ARRAY(sa.Integer()), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "preopen_ports")
    # ### end Alembic commands ###
