"""Fair Share GQL types package."""

from .common import (
    FairShareCalculationSnapshotGQL,
    FairShareSpecGQL,
    ResourceSlotEntryGQL,
    ResourceSlotGQL,
    ResourceWeightEntryInputGQL,
)
from .domain import (
    BulkUpsertDomainFairShareWeightInput,
    BulkUpsertDomainFairShareWeightPayload,
    DomainFairShareConnection,
    DomainFairShareDomainNestedFilter,
    DomainFairShareEdge,
    DomainFairShareFilter,
    DomainFairShareGQL,
    DomainFairShareOrderBy,
    DomainFairShareOrderField,
    DomainWeightInputItem,
    UpsertDomainFairShareWeightInput,
    UpsertDomainFairShareWeightPayload,
)
from .project import (
    BulkUpsertProjectFairShareWeightInput,
    BulkUpsertProjectFairShareWeightPayload,
    ProjectFairShareConnection,
    ProjectFairShareEdge,
    ProjectFairShareFilter,
    ProjectFairShareGQL,
    ProjectFairShareOrderBy,
    ProjectFairShareOrderField,
    ProjectFairShareProjectNestedFilter,
    ProjectFairShareTypeEnum,
    ProjectFairShareTypeEnumFilter,
    ProjectWeightInputItem,
    UpsertProjectFairShareWeightInput,
    UpsertProjectFairShareWeightPayload,
)
from .user import (
    BulkUpsertUserFairShareWeightInput,
    BulkUpsertUserFairShareWeightPayload,
    UpsertUserFairShareWeightInput,
    UpsertUserFairShareWeightPayload,
    UserFairShareConnection,
    UserFairShareEdge,
    UserFairShareFilter,
    UserFairShareGQL,
    UserFairShareOrderBy,
    UserFairShareOrderField,
    UserFairShareUserNestedFilter,
    UserWeightInputItem,
)

__all__ = [
    # Common
    "ResourceSlotEntryGQL",
    "ResourceSlotGQL",
    "ResourceWeightEntryInputGQL",
    "FairShareSpecGQL",
    "FairShareCalculationSnapshotGQL",
    # Domain
    "DomainFairShareGQL",
    "DomainFairShareConnection",
    "DomainFairShareDomainNestedFilter",
    "DomainFairShareEdge",
    "DomainFairShareFilter",
    "DomainFairShareOrderField",
    "DomainFairShareOrderBy",
    "UpsertDomainFairShareWeightInput",
    "UpsertDomainFairShareWeightPayload",
    "DomainWeightInputItem",
    "BulkUpsertDomainFairShareWeightInput",
    "BulkUpsertDomainFairShareWeightPayload",
    # Project
    "ProjectFairShareGQL",
    "ProjectFairShareConnection",
    "ProjectFairShareEdge",
    "ProjectFairShareFilter",
    "ProjectFairShareOrderField",
    "ProjectFairShareOrderBy",
    "ProjectFairShareProjectNestedFilter",
    "ProjectFairShareTypeEnum",
    "ProjectFairShareTypeEnumFilter",
    "UpsertProjectFairShareWeightInput",
    "UpsertProjectFairShareWeightPayload",
    "ProjectWeightInputItem",
    "BulkUpsertProjectFairShareWeightInput",
    "BulkUpsertProjectFairShareWeightPayload",
    # User
    "UserFairShareGQL",
    "UserFairShareConnection",
    "UserFairShareEdge",
    "UserFairShareFilter",
    "UserFairShareOrderField",
    "UserFairShareOrderBy",
    "UserFairShareUserNestedFilter",
    "UpsertUserFairShareWeightInput",
    "UpsertUserFairShareWeightPayload",
    "UserWeightInputItem",
    "BulkUpsertUserFairShareWeightInput",
    "BulkUpsertUserFairShareWeightPayload",
]
