"""Deployment policy action definitions."""

from .get_deployment_policy import (
    GetDeploymentPolicyAction,
    GetDeploymentPolicyActionResult,
)

__all__ = [
    "GetDeploymentPolicyAction",
    "GetDeploymentPolicyActionResult",
]
