"""Revision operations action definitions."""

from .activate_revision import (
    ActivateRevisionAction,
    ActivateRevisionActionResult,
)

__all__ = [
    "ActivateRevisionAction",
    "ActivateRevisionActionResult",
]
